"""
Command entry point.
"""

# pylint: disable=consider-using-f-string

import argparse
import json
import os
import time

from vidcrawler.spider import CRAWLER_MAP, crawl_video_sites

CRAWLERS = CRAWLER_MAP.keys()


def main() -> None:
    """Main function."""
    parser = argparse.ArgumentParser("vidcrawler")
    parser.add_argument("--input_crawl_json", type=str)
    parser.add_argument("--output_json", type=str)
    parser.add_argument("--singlethreaded", action="store_true")
    args = parser.parse_args()
    input_crawl_json = args.input_crawl_json or input("input_crawl_json: ")
    assert os.path.exists(input_crawl_json), f"{input_crawl_json} doesn't exist"
    output_json = args.output_json or input("output_json: ")
    with open(input_crawl_json, encoding="utf-8", mode="r") as filed:
        input_crawl_data = json.loads(filed.read())
    # Type check! We expect an array of tuples
    # List[Tuple[str, str, str]]
    tmp = []
    for tup in input_crawl_data:
        assert len(tup) == 3, "expected a List[Tuple[str, str, str]]"
        if tup[1] not in CRAWLERS:
            print(f"Unknown source: {tup[1]} (ignoring)")
            continue
        tmp.append(tup)
    input_crawl_data = tmp
    # Execute the crawl
    time_start = time.time()
    json_str: str = crawl_video_sites(input_crawl_data, use_threads=not args.singlethreaded)
    time_delta = time.time() - time_start
    print("\nTook %.1f seconds to fetch content\n" % time_delta)
    with open(output_json, encoding="utf-8", mode="w") as filed:
        filed.write(json_str)



"""
Utilities for reading and writing files.
"""

# pylint: disable=missing-function-docstring,too-many-arguments,too-many-locals,too-many-statements,line-too-long,invalid-name

import gzip
import json
import os
import sys
from datetime import datetime
from typing import Any, Dict, List

from .date import iso_fmt
from .video_info import VideoInfo


def write_gzip(out_path: str, content: bytes) -> None:
    assert out_path.endswith(".gz")
    with gzip.open(out_path, "wb") as f:
        f.write(content)


def write_utf8(out_path: str, content: str) -> None:
    out_path = os.path.normpath(out_path)
    dir_path: str = os.path.dirname(out_path)
    os.makedirs(dir_path, exist_ok=True)
    with open(out_path, "w", encoding="utf-8") as fd:
        fd.write(content)


def make_export_json(now: datetime, content: List[Any], network_name: str, telegram: str) -> Dict[str, Any]:
    # Type check, do we have a list of VideoInfo objects? If so convert to a list of dicts.
    if len(content) > 0:
        first = content[0]
        if isinstance(first, VideoInfo):
            content = VideoInfo.to_plain_list(content)
    return {
        "network_name": network_name,
        "__update_time": iso_fmt(now),
        "telegram": telegram,
        "content": content,
    }


def write_json(out_path: str, content: Any, also_gzip: bool = False, minify: bool = False, **kwargs) -> None:  # type: ignore
    """
    If zip is True, then the same json file will ALSO be written out
    with the extension *.zip and a single file inside labeled data.json.
    """
    assert "gzip" not in kwargs
    if not minify:
        kwargs["indent"] = 2
    out_utf8 = json.dumps(content, ensure_ascii=False, **kwargs)
    write_utf8(out_path, out_utf8)
    if also_gzip:
        if "indent" in kwargs:
            del kwargs["indent"]
        out_utf8 = json.dumps(content, ensure_ascii=False, **kwargs)
        write_gzip(out_path + ".gz", out_utf8.encode("utf-8"))


def read_utf8(in_path: str) -> str:
    in_path = os.path.normpath(in_path)
    with open(in_path, "r", encoding="utf-8") as fd:
        return fd.read()


def trap_win32_keyboard_interrupt_here() -> None:
    if sys.platform == "win32":  # Play better with ctrl-c
        import signal  # pylint: disable=import-outside-toplevel

        def raise_exc(a: Any, b: Any) -> None:
            raise KeyboardInterrupt("ctrl-c")

        signal.signal(signal.SIGINT, raise_exc)



# pylint: disable=missing-class-docstring

"""
Decorate string types with some classes.
"""


class VideoId(str):
    pass


class ChannelId(str):
    pass


class ChannelName(str):
    pass


class ChannelUrl(str):
    pass



# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return
"""
Fetcher for html
"""

import subprocess
import tempfile
from dataclasses import dataclass
from typing import Optional

import requests

try:
    subprocess.check_output("curl --version", shell=True)
    USE_CURL = True
except BaseException:  # pylint: disable=broad-except
    USE_CURL = False


@dataclass
class FetchResult:
    html: str
    status_code: int
    ok = property(lambda self: 200 <= self.status_code < 300)


def fetch_html_using_request_lib(
    url: str,
    timeout: Optional[int] = None,
    user_agent: Optional[str] = None,
) -> FetchResult:
    timeout = timeout or 10
    # Workaround for long request time on windows
    # see https://github.com/psf/requests/issues/4023
    headers = {"Connection": "close"}
    # headers = {"Connection": "close"}
    if user_agent:
        headers["User-Agent"] = user_agent
    resp = requests.get(url, timeout=timeout, params={}, headers=headers)
    resp.raise_for_status()
    return FetchResult(html=resp.text, status_code=resp.status_code)


def fetch_html_using_curl(url: str, timeout: Optional[int] = None) -> FetchResult:
    """Uses the curl library to fetch HTML and return HTML content and status code."""
    timeout = int(timeout or 10)
    # Create a temporary directory to store the response and the status code
    with tempfile.TemporaryDirectory() as temp_dir:
        # Define file paths within the temporary directory
        body_file_path = f"{temp_dir}/response_body.txt"
        status_code_file_path = f"{temp_dir}/status_code.txt"

        # Construct curl command to write body to a file and status code to another file
        command = f"curl --max-time {timeout} -s -o {body_file_path} -w '%{{http_code}}' -X GET {url} > {status_code_file_path}"
        subprocess.check_output(command, shell=True)

        # Read the status code from its file
        with open(status_code_file_path, encoding="utf-8", mode="r") as file:
            status_code = int(file.read().strip().replace("'", ""))

        # Read the response body from its file
        with open(body_file_path, encoding="utf-8", mode="r") as file:
            body = file.read()

    return FetchResult(html=body, status_code=status_code)


def fetch_html(url: str) -> FetchResult:
    if USE_CURL:
        return fetch_html_using_curl(url)
    else:
        return fetch_html_using_request_lib(url)



"""
Download a youtube video as an mp3.
"""

import os
import shutil
import subprocess
import tempfile
import warnings

from docker_run_cmd.api import docker_run
from static_ffmpeg import add_paths

FFMPEG_PATH_ADDED = False


def yt_dlp_download_mp3(url: str, outmp3: str) -> None:
    """Download the youtube video as an mp3."""
    global FFMPEG_PATH_ADDED  # pylint: disable=global-statement
    if not FFMPEG_PATH_ADDED:
        add_paths()
        FFMPEG_PATH_ADDED = True
    par_dir = os.path.dirname(outmp3)
    if par_dir:
        os.makedirs(par_dir, exist_ok=True)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = os.path.join(temp_dir, "temp.mp3")
        for _ in range(3):
            try:
                cmd_list: list[str] = []
                cmd_list += ["yt-dlp", url]
                is_youtube = "youtube.com" in url or "youtu.be" in url
                if is_youtube:
                    cmd_list += [
                        "-f",
                        "bestaudio",
                    ]
                cmd_list += [
                    "--extract-audio",
                    "--audio-format",
                    "mp3",
                    "--output",
                    temp_file,
                ]
                subprocess.run(cmd_list, check=True)
                shutil.copy(temp_file, outmp3)
                return
            except subprocess.CalledProcessError as cpe:
                print(f"Failed to download {url} as mp3: {cpe}")
                continue
        warnings.warn(f"Failed all attempts to download {url} as mp3.")


def docker_yt_dlp_download_mp3(url: str, outmp3: str) -> None:
    """Download the youtube video as an mp3."""
    here = os.path.abspath(os.path.dirname(__file__))
    dockerfile = os.path.join(here, "Dockerfile")
    dockerfile = os.path.abspath(dockerfile)
    assert os.path.exists(dockerfile), f"dockerfile {dockerfile} does not exist"
    with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
        cmd_args = [url, "-f", "bestaudio", "--extract-audio", "--audio-format", "mp3", "--output", "/host_dir/temp.mp3", "--update", "--no-geo-bypass"]
        docker_run(name="yt-dlp", dockerfile_or_url=dockerfile, cwd=temp_dir, cmd_list=cmd_args)
        shutil.copy(os.path.join(temp_dir, "temp.mp3"), outmp3)


def download_mp3(url: str, outmp3: str) -> None:
    """Download the youtube video as an mp3."""
    docker_yt_dlp = os.environ.get("USE_DOCKER_YT_DLP", "0") == "1"
    if docker_yt_dlp:
        return docker_yt_dlp_download_mp3(url, outmp3)
    return yt_dlp_download_mp3(url, outmp3)


def update_yt_dlp() -> None:
    docker_yt_dlp = os.environ.get("USE_DOCKER_YT_DLP", "0") == "1"
    if not docker_yt_dlp:
        return
    cmd_list = ["yt-dlp", "--update"]
    subprocess.run(cmd_list, check=True)


def unit_test() -> None:
    """Run the tests."""
    url = "https://www.youtube.com/watch?v=3Zl9puhwiyw"
    outmp3 = "tmp.mp3"
    download_mp3(url, outmp3)
    print(f"Downloaded {url} as {outmp3}")
    os.remove(outmp3)


if __name__ == "__main__":
    unit_test()



"""
Command line interface for fetching rumble channel
"""

import argparse
import json
import sys
from datetime import datetime
from pathlib import Path

from vidcrawler.date import parse_datetime
from vidcrawler.rumble import PartialVideo, fetch_rumble_channel_all_partial_result


def main() -> int:
    """Command line interface for fetching rumble channel."""
    parser = argparse.ArgumentParser(description="Fetch rumble channel.")
    parser.add_argument("--channel", help="Channel ID.", required=True)
    # add after date
    parser.add_argument("--after-date", help="Fetch videos after this date.", default=None)
    # add --output-json
    parser.add_argument("--output-json", help="Output to this file.", default=None)
    args = parser.parse_args()
    after_date: datetime | None = None
    if args.after_date:
        after_date = parse_datetime(args.after_date)
    vid_list: list[PartialVideo] = fetch_rumble_channel_all_partial_result(
        channel_name=args.channel,
        channel=args.channel,
        after=after_date,
    )
    vid_dict_list = [vid.to_dict() for vid in vid_list]
    json_data = json.dumps(vid_dict_list, indent=2)
    if args.output_json:
        # convert to json dict
        output_json = Path(args.output_json)
        output_json.write_text(json_data, encoding="utf-8")
    else:
        print(json_data)
    return 0


if __name__ == "__main__":
    sys.argv.append("--channel")
    sys.argv.append("PlandemicSeriesOfficial")
    sys.exit(main())



# pylint: disable=invalid-name,consider-using-f-string,fixme,missing-function-docstring,R0801,too-many-locals


"""
YouTube scraper.
"""


import datetime
import json
import os
import re
import subprocess
import sys
import tempfile
import time
import traceback
from typing import Any, List, Optional

import feedparser  # type: ignore
import requests  # type: ignore
from bs4 import BeautifulSoup  # type: ignore
from certifi import where
from keyvalue_sqlite import KeyValueSqlite as KeyValueDB  # type: ignore

from .date import iso8601_duration_as_seconds, iso_fmt, now_local
from .error import log_error
from .fetch_html import FetchResult, fetch_html
from .video_info import VideoInfo

where()  # This is to avoid a warning from requests about SSL certs.

_ENABLE_PROFILE_FETCH = False

HERE = os.path.dirname(__file__)
DB_YOUTUBE_CACHE = os.path.join(HERE, "cache", "youtube_cache.db")
os.makedirs(os.path.dirname(DB_YOUTUBE_CACHE), exist_ok=True)

# Retired. Lots of cost and only 10% of videos have subtitles.
# subtitle_fetcher_cache = YtSubtitleFetcherCache(DB_YOUTUBE_SUBTITLE)


def _get_cache(cache_path: str) -> KeyValueDB:
    return KeyValueDB(cache_path, "cached_youtube_video_attributes")


def _try_get_cached_duration(url: str, cache_path: Optional[str]) -> Optional[int]:
    if cache_path is None:
        return None
    cache = _get_cache(cache_path)
    cached_data = cache.get(url)
    if cached_data:
        duration = cached_data.get("duration_seconds")
        if duration is not None:
            return int(duration)
    return None


def _set_cached_duration(url: str, duration_secs: int, cache_path: Optional[str]) -> None:
    if cache_path is None:
        return
    cache = _get_cache(cache_path)
    cached_data = cache.get(url)
    if not cached_data:
        cached_data = {}
    cached_data.update({"duration_seconds": duration_secs})
    cache[url] = cached_data


def strfdelta(duration_seconds: int) -> str:
    if duration_seconds == 0:
        return "Live"
    td = datetime.timedelta(seconds=duration_seconds)
    hours, rem = divmod(td.seconds, 3600)
    mins, seconds = divmod(rem, 60)
    if hours == 0:
        return "%d:%02d" % (mins, seconds)
    return "%d:%02d:%02d" % (hours, mins, seconds)


def fetch_youtube_duration_str(url: str, cache_path: Optional[str]) -> str:
    sys.stdout.write("  Youtube visiting video %s\n" % url)
    cached_duration = _try_get_cached_duration(url, cache_path)
    if cached_duration:
        if cached_duration > -1:
            return strfdelta(cached_duration)
        return ""  # Gracefully handle error condition.
    try:
        fetch_result: FetchResult = fetch_html(url)
        html_doc = fetch_result.html
        assert html_doc, f"{__file__}: Could not fetch html doc from {url}"
        soup = BeautifulSoup(html_doc, "html.parser")
        dom = soup.find("meta", {"itemprop": "duration"})
        assert dom, f"{__file__}: Could not find duration in html doc in {url}"
        duration_str = dom.attrs["content"]
        duration_seconds = iso8601_duration_as_seconds(duration_str)
        _set_cached_duration(url, duration_seconds, cache_path)
        out = strfdelta(duration_seconds)
        return out
    except requests.exceptions.HTTPError as e:
        sys.stderr.write(f'{__file__} Error while processing {url} for duration because "{str(e)}"\n')
        _set_cached_duration(url, -1, cache_path)
        return ""
    except Exception as e:  # pylint: disable=broad-except
        # stack_trce = sys.exc_info()[2]
        stack_trce = traceback.format_exc()
        sys.stderr.write(f'{__file__} Error while processing {url} for duration because "{str(e)}"\n\n{stack_trce}')
        _set_cached_duration(url, -1, cache_path)
        return ""


def _fetch_youtube_channel_via_rss(channel_name: str, channel_id: str) -> List[VideoInfo]:
    url = "https://www.youtube.com/feeds/videos.xml?channel_id=" + channel_id
    sys.stdout.write(f"Youtube visiting {channel_name} ({url})\n")
    response: FetchResult = fetch_html(url)
    if response.status_code != 200:
        raise OSError(f"Could not fetch {url}")
    content = response.html
    if "was not found on this server" in content:  # TODO: Make less hacky.
        raise OSError(f"Could not fetch {url}")
    feed = feedparser.parse(content)
    output: List[VideoInfo] = []
    profile_picture = None
    for entry in feed.entries:
        views = entry.media_statistics["views"]
        if views == "0":  # Skip views with 0 as they have not be released yet.
            continue
        # Disable profile picture for now.
        if _ENABLE_PROFILE_FETCH and profile_picture is None:
            sys.stdout.write(f"  Youtube visiting {entry.link} to fetch profile picture.\n")
            # vid_html = fetch_html(entry.link)
            response = fetch_html(entry.link)
            vid_html = response.html
            if response.status_code != 200:
                sys.stdout.write(f"    Error - {entry.link} returned status code {response.status_code}\n")
            if vid_html is None:
                sys.stdout.write(f"    Error - vid_html from {entry.link} was None")
            else:
                try:
                    vid_info = parse_youtube_video(vid_html)
                    profile_picture = vid_info.get("profile_thumbnail", None)
                except AttributeError as err:
                    sys.stdout.write(f"Failed to get photo from {entry.link} because {err}.\n")
        now_datestr = iso_fmt(now_local())
        o = VideoInfo(
            channel_name=channel_name,
            title=entry.title,
            date_published=entry.published,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            channel_url=entry.author_detail.href,
            source="youtube.com",
            url=entry.link,
            duration="?",
            description=entry.summary,
            img_src=entry.media_thumbnail[0]["url"],
            iframe_src="https://youtube.com/embed/" + entry.yt_videoid,
            views=views,
            profile_img_src=profile_picture or "",
        )
        output.append(o)
    return output


def _fetch_youtube_channel_via_html(
    channel_name: str,
    channel_id: str,
    limit: int = -1,
) -> List[Any]:
    """SLOW! Also, Experimental - use if the videos.xml api is disabled"""

    # Step 1: using the channel_id fetch the channel_name.
    # TODO: this should be cached to avoid a fetch.
    url = f"https://www.youtube.com/channel/{channel_id}"
    sys.stdout.write(f"Youtube visiting {channel_name} ({url})\n")
    response: FetchResult = fetch_html(url)
    html_doc = response.html
    if not response.ok:
        raise OSError(f"Could not fetch {url}")
    pat = r'\"canonicalBaseUrl\":\"/c/([^\"]+)"'
    canonical_name = re.findall(pat, html_doc)[0]
    # Now the fetch the videos list from the channels canonical name
    # url.
    url = f"https://youtube.com/c/{canonical_name}/videos"
    response = fetch_html(url)
    chan_html_doc = response.html
    if not response.ok:
        raise OSError(f"Could not fetch {url}")
    # Inside a json file there is a bunch of videos that can be found
    # using the following regex.
    pat = r"{\"videoId\":\"([^\"]+)\","
    vid_ids = list(set(re.findall(pat, chan_html_doc)))
    vid_urls = [f"https://youtube.com/watch?v={id}" for id in vid_ids]
    output = []
    # Iterate through all videos and use the youtube-dl python command to
    # get a json dump of the file. Unfortunately this does not include the
    # very import publishing date time stamp. Another note is that with a
    # little bit of research it's not clear that there are alternative ways
    # to get exact timestamps from Youtube other than by videos.xml api, or
    # the youtube api which requires an audit of the app. Work arounds include
    # cross referencing the videos with those scraped from social blade or using
    # a paid service to get these videos.
    for i, vid_url in enumerate(vid_urls):
        if limit != -1 and i >= limit:
            break
        stdout = subprocess.check_output(f"youtube-dl {vid_url} -j", shell=True).decode("utf-8")
        data = json.loads(stdout)
        # Future:
        #  title = data['title']
        #  view_count = data['view_count']
        output.append(data)
    return output


def fetch_youtube_today(
    channel_name: str,
    channel_id: str,
    cache_path: Optional[str] = DB_YOUTUBE_CACHE,
    limit: int = -1,
) -> List[VideoInfo]:
    start_time = time.time()
    video_list = _fetch_youtube_channel_via_rss(channel_name, channel_id)
    if limit != -1:
        video_list = video_list[0:limit]
    vid: VideoInfo
    for vid in video_list:
        # Add in duration with per-video fetch
        duration_time = fetch_youtube_duration_str(vid.url, cache_path)
        vid.duration = duration_time
    delta_time = start_time - time.time()
    if delta_time > 15:
        sys.stdout.write(f"WARNING, youtube scraper took {int(delta_time)} seconds to complete\n")
    return video_list


def _test_fetch_youtube_video_info(url: str) -> None:
    sys.stdout.write("Youtube visiting video %s\n" % url)
    # html_doc = fetch_html(url)
    response: FetchResult = fetch_html(url)
    html_doc = response.html
    assert response.ok, f"{__file__}: Could not fetch html doc from {url}, because of status code {response.status_code}"
    assert html_doc, f"{__file__}: Could not fetch html doc from {url}"

    premiered_ago_matches = re.findall(r'\"Premiered (.{1,15}) ago"}', html_doc)
    premiered_in_progress = re.findall(r'\"Premiere in progress\. Started (.{1,15}) ago"}', html_doc)
    started_streaming_matches = re.findall(r"\"Started streaming (.{1,15}) ago\"", html_doc)
    video_is_private = '{"simpleText":"Private video"}' in html_doc
    unplayable = '"status":"UNPLAYABLE"' in html_doc
    # has_set_reminder = re.findall(r'\"simpleText\":\"Set reminder\"', html_doc)
    # has_set_reminder = '{"simpleText":"Set reminder"}' in html_doc

    stream_offline = '"status":"LIVE_STREAM_OFFLINE"' in html_doc

    print("  Testing %s" % url)
    print("    Premiered ago: %s" % premiered_ago_matches)
    print("    Premiered in progress started: %s" % premiered_in_progress)
    print("    Started streaming ago: %s" % started_streaming_matches)
    print("    Has stream_offline: %s" % stream_offline)
    print("    Is unplayable: %s" % unplayable)
    print("    Video is private: %s" % video_is_private)


# Still experimental
def parse_youtube_video(html_doc: str) -> dict:
    out: dict = {
        "date_published": "?",
        "views": "?",
        "profile_thumbnail": "?",
        "is_live": "?",
    }
    if '"isLiveNow":true' in html_doc:
        out["is_live"] = "True"
    else:
        out["is_live"] = "False"
    top_dom = BeautifulSoup(html_doc, "html.parser")  # type: ignore
    try:
        dom = top_dom.find("meta", {"itemprop": "startDate"})  # type: ignore
        out["date_published"] = str(dom.attrs["content"])
    except KeyError as err:
        log_error(str(err))
    except AttributeError as err:
        text = "NO STATUS"
        try:
            text = top_dom.text
        except Exception as err2:  # pylint: disable=broad-except
            print(f"{__file__}: Error: {err2}")
        sys.stdout.write(f"{__file__}: \ntop_dom.text: {text}\n\nError: {err}\n")
        raise
    try:
        dom = top_dom.find("meta", {"itemprop": "interactionCount"})  # type: ignore
        out["views"] = str(dom.attrs["content"])
    except KeyError as ke:
        log_error(str(ke))
    all_scripts = top_dom.find_all("script")
    needle_initial_data = "var ytInitialData = "
    # Needed for commented out code, below.
    # needle_player_response = "var ytInitialPlayerResponse = "
    for script in all_scripts:
        if not script.string:
            continue
        try:
            if script.string.startswith(needle_initial_data):
                img_urls = re.findall(r"\"(https://yt\d[^\"]+)\"", script.string)
                for img_url in img_urls:
                    if "s176" in img_url:  # For some reason this appears only in the thumbnail image.
                        out["profile_thumbnail"] = img_url
                        break
            # Video Details contains a lot of gems, but is disabled for now.
            # if script.string.startswith(needle_player_response):
            #     last_semi: int = script.string.rfind(";")
            #     json_str = script.string[
            #         len(needle_player_response) : last_semi
            #     ]
            #     json_data = json.loads(json_str)
            #     video_details = json_data["videoDetails"]
        except BaseException as be:  # pylint: disable=broad-except
            log_error("Error: " + str(be) + " for " + str(out))
    return out


def test_fetch_duration():
    url = "https://www.youtube.com/watch?v=UywxSWjzGaU"
    try:
        cache_path_file = tempfile.NamedTemporaryFile(suffix=".db", delete=False)  # pylint: disable=consider-using-with
        cache_path = cache_path_file.name
        cache_path_file.close()
        duration_str = fetch_youtube_duration_str(url, cache_path=cache_path)
        print("Discovered that the video at %s is %s" % (url, duration_str))
        duration_str = fetch_youtube_duration_str(url, cache_path=cache_path)
        print("Discovered that the video at %s is %s" % (url, duration_str))
    finally:
        os.remove(cache_path)


if __name__ == "__main__":
    # test_fetch_duration()
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=5nBqIK0mrFI")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=r5VjrVqgEU8")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=xfJsiDkbjg0")
    # Project veritas should not return anything
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=F7y19uNaYfE")
    # Upcoming film
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=yv7NFn95R0s")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=egRMAnwUQDQ")



# pylint: disable=missing-class-docstring,missing-function-docstring,missing-module-docstring

"""
Testing generic downloader
"""

import unittest

from vidcrawler.generic_downloader import fetch_all_videos_in_channel
from vidcrawler.types import ChannelId


class GenericDownloadTester(unittest.TestCase):
    def test_env(self):  # pylint: disable
        vidlist = fetch_all_videos_in_channel("youtube", ChannelId("UCiuTGTCkYrjVknhvMAICFjA"))
        self.assertGreater(len(vidlist), 0)


if __name__ == "__main__":
    unittest.main()



import json
import re
import subprocess
import warnings
from typing import Any

from vidcrawler.types import ChannelId, VideoId


def fetch_channel_info_ytdlp(video_url: str) -> dict[Any, Any]:
    """Fetch the info."""
    # yt-dlp -J "VIDEO_URL" > video_info.json
    cmd_list = [
        "yt-dlp",
        "-J",
        video_url,
    ]
    completed_proc = subprocess.run(cmd_list, capture_output=True, text=True, shell=True, check=True)
    if completed_proc.returncode != 0:
        stderr = completed_proc.stderr
        warnings.warn(f"Failed to run yt-dlp with args: {cmd_list}, stderr: {stderr}")
    lines: list[str] = []
    for line in completed_proc.stdout.splitlines():
        if line.startswith("OSError:"):
            continue
        lines.append(line)
    out = "\n".join(lines)
    data = json.loads(out)
    return data


def fetch_video_info(video_url: str) -> dict:
    cmd_list = [
        "yt-dlp",
        "-J",
        video_url,
    ]
    completed_proc = subprocess.run(cmd_list, capture_output=True, text=True, shell=True, check=True)
    if completed_proc.returncode != 0:
        stderr = completed_proc.stderr
        warnings.warn(f"Failed to run yt-dlp with args: {cmd_list}, stderr: {stderr}")
    lines: list[str] = []
    for line in completed_proc.stdout.splitlines():
        if line.startswith("OSError:"):
            continue
        lines.append(line)
    out = "\n".join(lines)
    data = json.loads(out)
    return data


def fetch_channel_url_ytdlp(video_url: str) -> str:
    """Fetch the info."""
    # yt-dlp -J "VIDEO_URL" > video_info.json
    cmd_list = [
        "yt-dlp",
        "--print",
        "channel_url",
        video_url,
    ]
    completed_proc = subprocess.run(cmd_list, capture_output=True, text=True, timeout=10, shell=True, check=True)
    if completed_proc.returncode != 0:
        stderr = completed_proc.stderr
        warnings.warn(f"Failed to run yt-dlp with args: {cmd_list}, stderr: {stderr}")
    lines = completed_proc.stdout.splitlines()
    out_lines: list[str] = []
    for line in lines:
        if line.startswith("OSError:"):  # happens on zach's machine
            continue
        out_lines.append(line)
    out = "\n".join(out_lines)
    return out


def fetch_channel_id_ytdlp(video_url: str) -> ChannelId:
    """Fetch the info."""
    url = fetch_channel_url_ytdlp(video_url)
    match = re.search(r"/channel/([^/]+)/?", url)
    if match:
        out: str = str(match.group(1))
        return ChannelId(out)
    raise RuntimeError(f"Could not find channel id in: {video_url} using yt-dlp.")


def fetch_videos_from_channel(channel_url: str) -> list[VideoId]:
    """Fetch the videos from a channel."""
    # yt-dlp -J "CHANNEL_URL" > channel_info.json
    # cmd = f'yt-dlp -i --get-id "https://www.youtube.com/channel/{channel_id}"'
    cmd_list = ["yt-dlp", "--print", "id", channel_url]
    cms_str = subprocess.list2cmdline(cmd_list)
    print(f"Running: {cms_str}")
    completed_proc = subprocess.run(
        cmd_list,
        capture_output=True,
        text=True,
        shell=True,
        check=True,
    )
    stdout = completed_proc.stdout
    lines = stdout.splitlines()
    out_channel_ids: list[VideoId] = []
    for line in lines:
        if line.startswith("OSError:"):  # happens on zach's machine
            continue
        if line.startswith("WARNING:"):
            warnings.warn(line)
            continue
        if line.startswith("ERROR:"):
            warnings.warn(line)
            continue
        out_channel_ids.append(VideoId(line))
    return out_channel_ids


def fetch_videos_from_youtube_channel(channel_id: str) -> list[VideoId]:
    """Fetch the videos from a youtube channel."""
    channel_url = f"https://www.youtube.com/channel/{channel_id}"
    return fetch_videos_from_channel(channel_url)



"""
    Scraper for bitchute
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name

import html
import re
import sys
from typing import List, Optional

import feedparser  # type: ignore
from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt, now_local
from .error import log_error

# bitchute is bombing out on CURL so switch to the request-lib get version.
from .fetch_html import fetch_html_using_request_lib
from .video_info import VideoInfo

_EMBED_BITCHUTE_PATT = r"/video/(.+)/"
USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36"


def fetch_html(url: str) -> str:
    return fetch_html_using_request_lib(url, user_agent=USER_AGENT).html


def parse_rss_url(html_doc: str) -> Optional[str]:
    top_dom = BeautifulSoup(html_doc, "html.parser")
    details_dom = top_dom.find("div", class_="details")
    if not details_dom:
        return None
    inner_dom = details_dom.find("p", class_="name")
    if not inner_dom:
        return None
    attrs = inner_dom.find("a")
    suffix = attrs["href"]
    rss_url: str = "https://www.bitchute.com/feeds/rss" + suffix
    return rss_url


def fetch_rss_url(channel_name: str, channel_id: str) -> Optional[str]:  # pylint: disable=unused-argument
    channel_url = "https://www.bitchute.com/channel/%s/" % channel_id
    html_doc = fetch_html(channel_url)
    return parse_rss_url(html_doc)


def parse_rss_feed(content: str) -> List[dict]:
    feed = feedparser.parse(content)
    channel_url = feed["feed"]["link"]
    # from pprint import pprint
    output: List[dict] = []
    for entry in feed.entries:
        date_published = entry["published"]
        summary = entry["summary"]
        title = entry["title"]
        id = entry["id"]  # pylint: disable=redefined-builtin
        try:
            type_ = entry["title_detail"]["type"]
            if type_ == "text/html":
                title = html.unescape(title)
        except BaseException as err:  # pylint: disable=broad-except
            log_error(f"Error while parsing because of {str(err)}")
        links = entry["links"]
        embed_link = links[0]["href"]
        img_link = links[1]["href"]
        id = re.findall(r".*/embed/([^\/]+)/", entry["link"])[0]
        obj = dict(  # pylint: disable=R1735
            title=title,
            date_published=date_published,
            alt_channel_url=channel_url,
            description=summary,
            img_src=img_link,
            iframe_src=embed_link,
            url="https://www.bitchute.com/video/%s" % id,
        )
        output.append(obj)
    return output


# type: ignore


def fetch_bitchute_today(channel_name: str, channel_id: str) -> List[VideoInfo]:
    output: List[VideoInfo] = []
    channel_url = "https://www.bitchute.com/channel/%s/" % channel_id
    sys.stdout.write("Bitchute visiting %s (%s)\n" % (channel_name, channel_url))
    html_doc = fetch_html(channel_url)
    rss_url = parse_rss_url(html_doc)
    date_published_map = {}
    if rss_url is None:
        sys.stderr.write("---- ERROR ---- Failed to parse rss_channel for %s\n" % channel_url)
    else:
        rss_content = fetch_html(rss_url)
        # These objects contain the video publishing date.
        rss_objects: List[dict] = parse_rss_feed(rss_content)
        rss_obj: dict
        for rss_obj in rss_objects:
            key = rss_obj["url"]
            date_published_map[key] = iso_fmt(rss_obj["date_published"])
    soup = BeautifulSoup(html_doc, "html.parser")
    vid_divs = soup.find_all(class_="channel-videos-container")
    skipped_vid_urls: List[str] = []
    for vd in vid_divs:
        title_dom = vd.find(class_="channel-videos-title")
        spa_dom = title_dom.find(class_="spa")
        title_text = spa_dom.text
        views_text = vd.find(class_="video-views").text.strip()
        video_src = spa_dom["href"]
        vid_id = re.findall(_EMBED_BITCHUTE_PATT, video_src)[0]
        iframe_src = "https://www.bitchute.com/embed/%s" % vid_id
        url = "https://www.bitchute.com/video/%s" % vid_id
        duration = vd.find(class_="video-duration").text  # type: ignore
        # poster_dom = plyr__poster
        poster_dom = vd.find(class_="channel-videos-image")
        img_src = poster_dom.find("img")["data-src"]
        now_datestr = now_local().isoformat()
        date_published = date_published_map.get(url)
        if date_published is None:
            skipped_vid_urls.append(url)
            continue
        o = VideoInfo(
            channel_name=channel_name,
            source="bitchute.com",
            date_published=date_published,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            url=url,
            channel_url=channel_url,
            title=title_text,
            duration=duration,
            description="TODO",
            img_src=img_src,
            iframe_src=iframe_src,
            views=views_text,
            profile_img_src="",
        )
        output.append(o)
    if skipped_vid_urls:
        sys.stdout.write(f"{__file__}: Skipping {len(skipped_vid_urls)} videos from {channel_name} because no data from rss.\n")
    return output



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import unittest

from vidcrawler.brighteon import fetch_brighteon_today


class BrighteonScraperTester(unittest.TestCase):
    def test_fetch_brighteon_today(self):
        vid_list = fetch_brighteon_today(channel_name="hrreport", channel="hrreport")
        self.assertIsNotNone(vid_list)
        self.assertGreater(len(vid_list), 0)

        for vid in vid_list:
            self.assertIsNotNone(vid)
            # assert that the title and other elements are valid
            self.assertGreater(len(vid.title), 0)
            self.assertGreater(len(vid.url), 0)
            self.assertGreater(len(vid.img_src), 0)
            self.assertGreater(len(vid.iframe_src), 0)
            self.assertGreater(len(vid.description), 0)
            # duration should not be ?
            self.assertNotEqual(vid.duration, "?")
            self.assertGreater(len(vid.duration), 0)
            self.assertGreater(len(vid.date_published), 0)
            self.assertGreater(len(vid.date_discovered), 0)
            self.assertGreater(len(vid.date_lastupdated), 0)
            self.assertGreater(len(vid.channel_name), 0)
            self.assertGreater(len(vid.source), 0)
            self.assertGreater(len(vid.channel_url), 0)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=R0801

import argparse
import os
import sys

from vidcrawler.library import Library, VidEntry
from vidcrawler.rumble import PartialVideo, fetch_rumble_channel_all_partial_result


def _update_library(outdir: str, channel_name: str) -> Library:
    # channel_url = f"https://www.brighteon.com/channels/{channel_name}"
    library_json = os.path.join(outdir, "library.json")
    videos: list[PartialVideo] = fetch_rumble_channel_all_partial_result(
        channel_name=channel_name,
        channel=channel_name,
        after=None,
    )
    urls: list[VidEntry] = [VidEntry(url=vid.url, title=vid.title, date=vid.date) for vid in videos]
    library = Library(library_json)
    library.merge(urls)
    return library


def main() -> int:
    """Main function."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--channel-name",
        type=str,
        help="URL slug of the channel, example: hrreport",
        required=True,
    )
    parser.add_argument("--output", type=str, help="Output directory", required=True)
    # full-scan
    parser.add_argument("--skip-download", action="store_true", help="Skip downloading")
    args = parser.parse_args()
    outdir = args.output
    channel = args.channel_name

    library = _update_library(outdir, channel)
    print(f"Updated library {library.library_json_path}")
    if not args.skip_download:
        library.download_missing()
    return 0


def unit_test() -> int:
    """Run the tests."""
    sys.argv.append("--channel-name")
    sys.argv.append("PlandemicSeriesOfficial")
    sys.argv.append("--output")
    sys.argv.append("tmp3")
    sys.argv.append("--skip-download")
    main()
    return 0


if __name__ == "__main__":
    sys.exit(unit_test())



# pylint: disable=all
# flake8: noqa

import unittest

from vidcrawler.youtube_bot import (
    VidEntry,
    fetch_all_sources,
    fetch_all_vids,
    parse_youtube_videos,
)

URL_SILVERGURU = "https://www.youtube.com/@silverguru/videos"


class YouTubeBotTester(unittest.TestCase):
    def test_fetch_sources(self) -> None:
        sources = list(fetch_all_sources(URL_SILVERGURU, limit=1))
        print("sources:")
        # for source in sources:
        #    print(f"  {source}")
        vids: list[VidEntry] = parse_youtube_videos(sources)
        self.assertGreater(len(vids), 0)
        print("vids:")
        for vid in vids:
            print(f"  {vid}")

    def test_fetch_videos(self) -> None:
        # vids = fetch_all_vids(URL_SILVERGURU)
        sources: list[VidEntry] = fetch_all_vids(URL_SILVERGURU, limit=4)
        print("vids:")
        for vid in sources:
            print(f"  {vid}")
        self.assertGreater(len(sources), 0)


if __name__ == "__main__":
    unittest.main()



"""
Rumble scrapper.
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return,fixme,too-many-branches,too-many-statements

import sys
import traceback
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import List

from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt, now_local, timestamp_to_iso8601
from .fetch_html import FetchResult
from .fetch_html import fetch_html_using_curl as fetch_html
from .video_info import VideoInfo
from .ytdlp import fetch_video_info


@dataclass
class RumbleResponse:
    html_doc: str
    channel_url: str
    fetch_result: FetchResult


def fetch_rumble_channel_url(channel_url: str) -> RumbleResponse:
    sys.stdout.write(f"Rumble visiting {channel_url}\n")
    # html_doc = fetch_html(channel_url)
    fetch_result: FetchResult = fetch_html(channel_url)
    # return fetch_result
    return RumbleResponse(fetch_result.html, channel_url, fetch_result)


def fetch_rumble(channel: str) -> RumbleResponse:
    html_doc: str = ""
    channel_url: str = "https://rumble.com/c/%s" % channel
    url: str = "https://rumble.com/c/%s?date=this-month" % channel
    try:
        sys.stdout.write("Rumble visiting %s (%s)\n" % (channel, url))
        # html_doc = fetch_rumble_channel_url(url)
        response: RumbleResponse = fetch_rumble_channel_url(url)
        if response.fetch_result.ok:
            html_doc = response.fetch_result.html
        return RumbleResponse(html_doc, channel_url, response.fetch_result)

    except KeyboardInterrupt:
        raise
    except SystemExit:  # pylint: disable=try-except-raise
        raise
    except Exception as err:  # pylint: disable=broad-except
        warnings.warn(f"Error fetching rumble channel {channel}: {err}")
        channel_url = "https://rumble.com/user/%s" % channel
        url = "https://rumble.com/user/%s" % channel
        sys.stdout.write(f"Rumble alt visiting {channel} ({url})\n")
        # html_doc = fetch_html(url)
        fetch_result = fetch_html(url)
        if fetch_result.ok:
            html_doc = fetch_result.html
        else:
            warnings.warn(f"Failed to fetch {url}")
        raise


def rumble_video_id_to_embed_url(video_id: str) -> str:
    return f"https://rumble.com/embed/{video_id}"


@dataclass
class PartialVideo:
    url: str
    title: str
    duration: str
    videoid: str
    channel_url: str
    channel_name: str
    date: datetime

    def to_dict(self) -> dict:
        """Returns a json serializable dictionary."""
        return {
            "url": self.url,
            "duration": self.duration,
            "videoid": self.videoid,
            "channel_url": self.channel_url,
            "channel_name": self.channel_name,
            "date": self.date.isoformat(),
        }


def fetch_rumble_channel_today_legacy(channel_name: str, channel: str) -> List[VideoInfo]:
    # TODO: Fine grained try...catch blocks.
    output: List[VideoInfo] = []
    html_doc: str = ""
    channel_url: str = ""
    # html_doc, channel_url = fetch_rumble(channel)
    fetch_response: RumbleResponse = fetch_rumble(channel)
    html_doc = fetch_response.html_doc
    channel_url = fetch_response.channel_url
    fetch_response = fetch_rumble_channel_url(channel_url)
    if not fetch_response.fetch_result.ok:
        warnings.warn(f"Failed to fetch {channel_url}")
        return []
    soup = BeautifulSoup(html_doc, "html.parser")
    for article in soup.find_all("div", class_="videostream thumbnail__grid--item"):
        try:
            article_duration_dom = article.find(class_="videostream__status--duration")
            duration = article_duration_dom.get_text().strip()
            vid_src_suffix = article.find("a", class_="videostream__link link")["href"]
            vid_src = "https://rumble.com%s" % vid_src_suffix
            sys.stdout.write("  visiting video %s (%s)\n" % (channel, vid_src))
            video_obj = fetch_video_info(vid_src)
            video_id = video_obj["id"]
            title = video_obj["fulltitle"]
            iframe_src = rumble_video_id_to_embed_url(video_id)
            desc_text = video_obj["description"]
            publish_timestamp = float(video_obj["timestamp"])
            publish_date = timestamp_to_iso8601(publish_timestamp)
            views = str(video_obj["view_count"])
            thumbnails: list[dict[str, str]] = video_obj["thumbnails"]
            img_src = thumbnails[0]["url"]
            now_datestr = iso_fmt(now_local())
            o = VideoInfo(
                channel_name=channel_name,
                source="rumble.com",
                date_published=publish_date,
                date_discovered=now_datestr,
                date_lastupdated=now_datestr,
                url=vid_src,
                channel_url=channel_url,
                title=title,
                duration=duration,
                description=desc_text,
                img_src=img_src,
                iframe_src=iframe_src,
                views=views,
                profile_img_src="",
            )
            output.append(o)
        except BaseException as e:  # pylint: disable=broad-except
            s = "".join(traceback.format_exception(None, e, e.__traceback__))
            sys.stdout.write("Error: %s\nCould not parse\n%s\n\n" % (str(s), str(article)))
    return output


def fetch_rumble_channel_today_partial_result(channel_name: str, channel: str) -> list[PartialVideo]:
    out: List[PartialVideo] = []
    html_doc: str = ""
    channel_url: str = ""
    # html_doc, channel_url = fetch_rumble(channel)
    response: RumbleResponse = fetch_rumble(channel)
    html_doc = response.html_doc
    channel_url = response.channel_url
    fetch_response = fetch_rumble_channel_url(channel_url)
    if not fetch_response.fetch_result.ok:
        warnings.warn(f"Failed to fetch {channel_url}")
        return []

    soup = BeautifulSoup(html_doc, "html.parser")
    for article in soup.find_all("div", class_="videostream thumbnail__grid--item"):
        try:
            article_duration_dom = article.find(class_="videostream__status--duration")
            duration = article_duration_dom.get_text().strip()
            vid_src_suffix = article.find("a", class_="videostream__link link")["href"]
            vid_src = "https://rumble.com%s" % vid_src_suffix
            dom_vid_status = article.find("div", class_="videostream__status")
            if dom_vid_status is not None:
                fuzzy_date = dom_vid_status.get_text().strip()
            else:
                fuzzy_date = ""
            title = parse_title(article)
            date: datetime = parse_fuzzy_date(fuzzy_date)
            partial_video: PartialVideo = PartialVideo(
                url=vid_src,
                title=title,
                duration=duration,
                videoid="",
                channel_url=channel_url,
                channel_name=channel_name,
                date=date,
            )
            out.append(partial_video)
        except BaseException as e:  # pylint: disable=broad-except
            s = "".join(traceback.format_exception(None, e, e.__traceback__))
            sys.stdout.write("Error: %s\nCould not parse\n%s\n\n" % (str(s), str(article)))
    return out


def parse_duration(article_soup: BeautifulSoup) -> str:
    dom_duration = article_soup.find("div", class_="videostream__status--duration")
    if dom_duration is not None:
        duration = dom_duration.get_text().strip()
        return duration
    dom_live = article_soup.find("a", class_="videostream__status--live")
    if dom_live is not None:
        if dom_live.get_text().strip().upper() == "LIVE":
            return "LIVE"
    dom_dvr = article_soup.find("a", class_="videostream__status--dvr")
    if dom_dvr is not None:
        if dom_dvr.get_text().strip().upper() == "DVR":
            return "DVR"
    raise ValueError("Could not find duration")


def parse_title(article_soup: BeautifulSoup) -> str:
    class_name = "thumbnail__title"
    dom_title = article_soup.find(class_=class_name)
    if dom_title is not None:
        title = dom_title.get_text().strip()
        return title
    raise ValueError(f"Could not find title with class {class_name}")


def parse_date(article_soup: BeautifulSoup) -> str:
    # videostream__data--item videostream__date
    dom_date = article_soup.find(class_="videostream__data--item videostream__date")
    if dom_date is not None:
        # does text attribute exist?
        if hasattr(dom_date, "title"):
            date = dom_date["title"]
            if date:
                return date
        return dom_date.get_text().strip()
    warnings.warn("Could not find date")
    return ""


def get_channel_url(channel: str, page_num: int, is_user_channel: bool) -> str:
    # only return a page if the page number is greater than 1
    # if page_num > 1:
    #    if is_user_channel:
    #        return f"https://rumble.com/user/{channel}?page={page_num}"
    #    return f"https://rumble.com/c/{channel}?page={page_num}"
    # if is_user_channel:
    #    return f"https://rumble.com/user/{channel}"
    # return f"https://rumble.com/c/{channel}"
    # let's simplify this
    if is_user_channel:
        base_url = f"https://rumble.com/user/{channel}"
    else:
        base_url = f"https://rumble.com/c/{channel}"
    if page_num > 1:
        return f"{base_url}?page={page_num}"
    return base_url


def parse_fuzzy_date(datestr: str) -> datetime:
    """Parses strings like 'November 6, 2023' into datetime objects."""
    return datetime.strptime(datestr, "%B %d, %Y")


def fetch_rumble_channel_all_partial_result(channel_name: str, channel: str, after: datetime | None = None) -> list[PartialVideo]:
    out: List[PartialVideo] = []
    page = 1
    is_user_channel = False
    test_url = get_channel_url(channel, page, is_user_channel)
    fetch_response = fetch_html(test_url)
    if not fetch_response.ok:
        is_user_channel = True
        # now assert they can be reached
        test_url = get_channel_url(channel, page, is_user_channel)
        fetch_response = fetch_html(test_url)
        if not fetch_response.ok:
            raise ValueError(f"Could not find channel or user {channel}")
    while True:
        try:
            curr_page = page
            page += 1
            html_doc: str = ""
            current_channel_url = get_channel_url(channel, curr_page, is_user_channel)
            # html_doc = fetch_rumble_channel_url(current_channel_url)
            fetch_result: FetchResult = fetch_html(current_channel_url)
            if fetch_result.ok:
                html_doc = fetch_result.html
            else:
                status_code = fetch_result.status_code
                if status_code == 404:
                    break  # expected result when we've reached the end of the channel.
                warnings.warn(f"Failed to fetch {current_channel_url}")
                break
            soup = BeautifulSoup(html_doc, "html.parser")
            for article in soup.find_all("div", class_="videostream thumbnail__grid--item"):
                try:
                    duration = parse_duration(article)
                    vid_src_suffix = article.find("a", class_="videostream__link link")["href"]
                    vid_src = f"https://rumble.com{vid_src_suffix}"
                    fuzzy_date = parse_date(article)
                    date = parse_fuzzy_date(fuzzy_date)
                    title = parse_title(article)
                    if after is not None and date < after:
                        continue
                    videoid = vid_src.split("/")[-1]
                    videoid = videoid.split("-")[0]
                    partial_video: PartialVideo = PartialVideo(
                        url=vid_src,
                        title=title,
                        duration=duration,
                        videoid=videoid,
                        channel_url=current_channel_url,
                        channel_name=channel_name,
                        date=date,
                    )
                    out.append(partial_video)
                except BaseException as e:  # pylint: disable=broad-except
                    s = "".join(traceback.format_exception(None, e, e.__traceback__))
                    sys.stdout.write("Error: %s\nCould not parse\n%s\n\n" % (str(s), str(article)))
        except KeyboardInterrupt:
            raise
        except SystemExit:  # pylint: disable=try-except-raise
            raise
        except Exception as err:  # pylint: disable=broad-except
            warnings.warn(f"Error fetching rumble channel {channel}: {err}")
    return out


def resolve(rumble_partial: PartialVideo) -> VideoInfo:
    vid_src = rumble_partial.url
    sys.stdout.write("  visiting video %s (%s)\n" % (rumble_partial.channel_name, vid_src))
    video_obj = fetch_video_info(vid_src)
    video_id = video_obj["id"]
    title = video_obj["fulltitle"]
    iframe_src = rumble_video_id_to_embed_url(video_id)
    try:
        desc_text = video_obj["description"]
    except KeyError:
        warnings.warn(f"Could not find description for {vid_src}")
        desc_text = ""
    try:
        publish_timestamp = float(video_obj["timestamp"])
    except KeyError:
        warnings.warn(f"Could not find timestamp for {vid_src}")
        publish_timestamp = 0
    publish_date = timestamp_to_iso8601(publish_timestamp)
    views = str(video_obj["view_count"])
    thumbnails: list[dict[str, str]] = video_obj["thumbnails"]
    img_src = thumbnails[0]["url"]
    now_datestr = iso_fmt(now_local())
    o = VideoInfo(
        channel_name=rumble_partial.channel_name,
        source="rumble.com",
        date_published=publish_date,
        date_discovered=now_datestr,
        date_lastupdated=now_datestr,
        url=vid_src,
        channel_url=rumble_partial.channel_url,
        title=title,
        duration=rumble_partial.duration,
        description=desc_text,
        img_src=img_src,
        iframe_src=iframe_src,
        views=views,
        profile_img_src="",
    )
    return o


def fetch_rumble_channel_today(channel_name: str, channel: str) -> List[VideoInfo]:
    # use the partial result, then full resolve
    partial_result = fetch_rumble_channel_today_partial_result(channel_name, channel)
    output: List[VideoInfo] = []
    for partial in partial_result:
        vinfo = resolve(partial)
        output.append(vinfo)
    return output


def interactive_test() -> None:
    sys.stdout.write("Rumble user: ")
    name: str = input()
    rslt = fetch_rumble_channel_today(channel_name=name, channel=name)
    print(rslt)


if __name__ == "__main__":
    interactive_test()



# flake8: noqa=W293
# pylint: disable=line-too-long,missing-class-docstring,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,redefined-builtin

"""
Test script for opening a youtube channel and getting the latest videos.
"""


from vidcrawler.types import ChannelId, ChannelUrl, VideoId
from vidcrawler.ytdlp import fetch_videos_from_channel


def to_rumble_channel_url(id: ChannelId) -> ChannelUrl:
    url: str = f"https://rumble.com/c/{id}/videos"
    return ChannelUrl(url)


def to_youtube_channel_url(id: ChannelId) -> ChannelUrl:
    url: str = f"https://www.youtube.com/channel/{id}"
    return ChannelUrl(url)


def fetch_all_videos_in_rumble_channel(channel_id: ChannelId) -> list[VideoId]:
    url: ChannelUrl = to_rumble_channel_url(channel_id)
    vids: list[VideoId] = fetch_videos_from_channel(url)
    return vids


def fetch_all_videos_in_youtube_channel(channel_id: ChannelId) -> list[VideoId]:
    url: ChannelUrl = to_youtube_channel_url(channel_id)
    vids: list[VideoId] = fetch_videos_from_channel(url)
    return vids


def fetch_all_videos_in_channel(channel_type: str, channel_id: ChannelId) -> list[VideoId]:
    if channel_type == "youtube":
        fetcher = fetch_all_videos_in_youtube_channel
    elif channel_type == "rumble":
        fetcher = fetch_all_videos_in_rumble_channel
    else:
        raise ValueError(f"Unknown channel_type: {channel_type}")
    out: list[VideoId] = fetcher(channel_id)
    return out


def unit_test() -> int:
    channel_id = ChannelId("UCiuTGTCkYrjVknhvMAICFjA")
    vidlist = fetch_all_videos_in_channel("youtube", channel_id)
    print(f"Found {len(vidlist)} videos.")
    for vid in vidlist:
        print(f"  {vid}")
    return 0


def main() -> int:
    return unit_test()


if __name__ == "__main__":
    main()



"""
    Scraper for Brighteon.com
"""

import re

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from typing import List

import requests
from bs4 import BeautifulSoup  # type: ignore
from feedparser import FeedParserDict, parse
from PIL import Image  # type: ignore

from .date import iso_fmt, now_local

# from .fetch_html import fetch_html_using_request_lib as fetch_html
from .video_info import VideoInfo

# https://www.brighteon.com/api-v3/channels/hrreport/rss/rss.xml


def get_rss_url(channel: str) -> str:
    return f"https://www.brighteon.com/api-v3/channels/{channel}/rss/rss.xml"


def parse_thumbnail_url(soup: BeautifulSoup) -> str:
    img = soup.find("img")
    if img:
        if "thumbnail" in img["src"]:
            return img["src"]
    raise ValueError("Could not find thumbnail image")


def parse_entry(url: str, channel_name: str, entry: FeedParserDict) -> VideoInfo:
    created_time = entry.published
    iframe_url = "https://www.brighteon.com/embed%s" % entry.id
    video_url = entry.link
    title = entry.title
    summary_html = entry.summary
    summary_soup = BeautifulSoup(summary_html, "html.parser")
    image_src = parse_thumbnail_url(summary_soup)
    # print(summary_html)
    # image_src = entry.media_thumbnail[0]["url"]
    view_count = "?"
    duration = "?"
    now_datestr = iso_fmt(now_local())
    vid = VideoInfo(
        channel_name=channel_name,
        source="brighteon.com",
        date_published=created_time,
        date_discovered=now_datestr,
        date_lastupdated=now_datestr,
        url=video_url,
        channel_url=url,
        title=title,
        duration=duration,
        description=summary_html,
        img_src=image_src,
        iframe_src=iframe_url,
        views=view_count,
        profile_img_src="",
    )
    return vid


def fetch_image_size(vid: VideoInfo) -> None:
    """fetches the height and width by downloading the image"""
    try:
        url: str = vid.img_src
        response = requests.get(url, stream=True, timeout=10)
        response.raw.decode_content = True
        img = Image.open(response.raw)
        vid.img_width = img.width
        vid.img_height = img.height
        vid.img_status = 200
    except Exception as verr:  # pylint: disable=broad-except
        sys.stderr.write(f"Error fetching image size: {verr} during {url}\n")


def fetch_views_and_duration(vid: VideoInfo) -> None:
    url = vid.url
    response = requests.get(url, timeout=10)
    text = response.text
    # pattern is like "3798 views"

    pattern = re.compile(r"(\d+) views")
    match = pattern.search(text)
    if match:
        out = str(match.group(1))
        vid.views = out
    else:
        vid.views = "?"

    # "duration":"22:27"
    # also "duration":"02:34:18"
    pattern = re.compile(r'"duration":"(\d+:\d+:\d+|\d+:\d+)"')
    match = pattern.search(text)
    if match:
        out = str(match.group(1))
        vid.duration = out
    else:
        vid.duration = "?"


def fetch_brighteon_today(channel_name: str, channel: str) -> List[VideoInfo]:
    # switching to rss feed style
    output: List[VideoInfo] = []
    url = get_rss_url(channel)
    sys.stdout.write("Brighteon visiting %s (%s)\n" % (channel, url))
    feed = parse(url)
    entry: FeedParserDict
    for entry in feed.entries:
        try:
            vid = parse_entry(url, channel_name, entry)
            output.append(vid)
        except Exception as verr:  # pylint: disable=broad-except
            sys.stderr.write(f"Error parsing entry: {verr} during {entry}\n")

    with ThreadPoolExecutor(max_workers=8) as executor:
        fetch_size = partial(fetch_image_size)
        futures = [executor.submit(fetch_size, vid) for vid in output]
        for future in as_completed(futures):
            future.result()  # This will raise any exceptions that occurred during execution

    # now bulk fetch the views
    with ThreadPoolExecutor(max_workers=8) as executor:
        fetch_views_partial = partial(fetch_views_and_duration)
        futures = [executor.submit(fetch_views_partial, vid) for vid in output]
        for future in as_completed(futures):
            future.result()  # This will raise any exceptions that occurred during execution

    return output



"""
Testing for io
"""

# pylint: disable=line-too-long,missing-function-docstring,missing-class-docstring,super-with-arguments,invalid-name,R0801

import os
import pathlib
import shutil
import tempfile
import traceback
import unittest
import warnings

from vidcrawler.library import Library

HERE = pathlib.Path(__file__).parent.absolute()
BAD_LIBRARY_JSON_PATH = os.path.join(HERE, "bad_library.json")
assert os.path.exists(BAD_LIBRARY_JSON_PATH), f"bad_library.json {BAD_LIBRARY_JSON_PATH} does not exist"


class BadVideoMp3DownloadTester(unittest.TestCase):

    def test_create_tmp_dir(self) -> None:
        prev_dir = os.getcwd()
        temp_dir = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        try:
            os.chdir(temp_dir.name)
            basename = os.path.basename(BAD_LIBRARY_JSON_PATH)
            shutil.copy(BAD_LIBRARY_JSON_PATH, ".")
            library = Library(basename)
            library.download_missing()
        except Exception as e:  # pylint: disable=broad-except
            traceback.print_exc()
            self.fail(f"Failed to download missing videos: {e}")
        finally:
            try:
                os.chdir(prev_dir)
                temp_dir.cleanup()
            except Exception as e:  # pylint: disable=broad-except
                warnings.warn(f"Failed to cleanup temp_dir: {e}")


if __name__ == "__main__":
    unittest.main()



"""
Spotify scraper
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return,fixme

import sys
from datetime import datetime
from typing import List

from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt
from .fetch_html import fetch_html_using_request_lib as fetch_html
from .video_info import VideoInfo

_TIMEOUT_EPISODE = 20  # Wow these can take a long time.

# TODO: Allow spotify scraper to access the database and check to see if
# the episode is already there.


def now_local() -> datetime:
    """Patch to allow code to work."""
    return datetime.now()


def fetch_spotify_today(channel_name: str, channel: str) -> List[VideoInfo]:
    output: List[VideoInfo] = []
    now_datestr = iso_fmt(now_local())
    channel_url = f"https://open.spotify.com/show/{channel}"
    sys.stdout.write(f"Spotify crawler visiting {channel_name} ({channel_url})\n")
    html_doc = fetch_html(channel_url).html
    html_dom = BeautifulSoup(html_doc, "html.parser")
    music_doms = html_dom.findAll("meta", {"name": "music:song"})
    episode_urls = [e.attrs["content"] for e in music_doms]
    for episide_url in episode_urls:
        sys.stdout.write(f"  Spotify crawler visiting episode {episide_url}\n")
        episode_html = fetch_html(episide_url, timeout=_TIMEOUT_EPISODE)
        episode_dom = BeautifulSoup(episode_html, "html.parser")

        def extract_meta_property(name: str) -> str:
            dom = episode_dom.find("meta", {"name": name})  # pylint: disable=cell-var-from-loop
            return str(dom["content"])

        img_url = extract_meta_property("og:image")
        title = extract_meta_property("og:title")
        description = extract_meta_property("og:description")
        release_date = extract_meta_property("music:release_date")
        duration = extract_meta_property("music:duration")
        url = extract_meta_property("og:url")
        # example:
        #   "4rJBoD4BeeYFd8eLhrDRM" is extracted from
        #   "https://open.spotify.com/episode/4rJBoD4BeeYFd8eLhrDRM"
        spotify_id = url.split("episode")[1].replace("/", "")
        title = extract_meta_property("og:title")
        vid_url = f"https://open.spotify.com/episode/{spotify_id}"
        iframe_url = f"https://open.spotify.com/embed/episode/{spotify_id}"
        vid: VideoInfo = VideoInfo(
            channel_name=channel_name,
            source="spotify.com",
            date_published=release_date,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            url=vid_url,
            channel_url=channel_url,
            title=title,
            duration=duration,
            description=description,
            img_src=img_url,
            iframe_src=iframe_url,
            views="?",
            profile_img_src="",
        )
        output.append(vid)
    return output


def unit_test() -> None:
    fetch_spotify_today(channel_name="Joe Rogan", channel="4rOoJ6Egrf8K2IrywzwOMk")


if __name__ == "__main__":
    # Unit test
    unit_test()



"""
Odysee scrapper.
"""

# pylint: disable=invalid-name

from typing import List

import feedparser  # type: ignore
import requests  # type: ignore

from .date import iso_fmt, now_local
from .video_info import VideoInfo

_TIMEOUT = 10


def _parse_rss_entry(entry: feedparser.util.FeedParserDict) -> VideoInfo:
    """Parses an rss entry and outputs a VideoInfo object."""
    title = entry.title
    link = entry.link
    links = entry.links
    published = entry.published
    thumbnail_img = ""
    for lnk in links:
        if lnk.type.lower().startswith("image/"):
            thumbnail_img = lnk.href
            break
    description = getattr(entry, "description", "")
    o = VideoInfo(
        channel_name="",
        title=title,
        date_published=published,
        date_discovered="",
        date_lastupdated="",
        source="odysee.com",
        url=link,
        duration="?",  # No duration, suprisingly.
        description=description,
        img_src=thumbnail_img,
        iframe_src="",
        views="?",
        profile_img_src=thumbnail_img,
    )
    return o


def fetch_odysee_today(channel_name: str, channel: str) -> List[VideoInfo]:
    """Fetches the latest videos from odysee.com."""
    url: str = f"https://lbryfeed.melroy.org/channel/odysee/{channel}"
    channel_url: str = f"https://odysee.com/@{channel}"
    resp = requests.get(url, timeout=_TIMEOUT)
    now_str: str = iso_fmt(now_local())
    resp.raise_for_status()
    feed = feedparser.parse(resp.text)
    out: List[VideoInfo] = []
    for entry in feed.entries:
        vo = _parse_rss_entry(entry)
        vo.channel_name = channel_name
        vo.channel_url = channel_url
        vo.date_discovered = now_str
        vo.date_lastupdated = now_str
        if vo:
            out.append(vo)
    return out
